#include <stdio.h>
#include <windows.h>
#include "privilege.h"

#include <WindowsX.h>
#include <commctrl.h>
#include <prsht.h>
#include <shellapi.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <objbase.h>

void print_help() {
    printf("Usage:\n"
           "  xln [<switch> [<keys>]] <target> <link name>\n"
           "Switches:\n"
           "  -h       NTFS hard link (assumed by default)\n"
           "  -n       NTFS reparse point\n"
           "  -w       Windows Shell link\n"
           "Shell link keys (for .LNK files):\n"
           "  -d  <description>\n"
           "  -wd <working directory>\n"
           "  -a  <command line arguments>\n"
           "  -r  assume target path is relative\n"
           "NTFS reparse point keys:\n"
//           "  -s  make symbolik link\n"
           "  -b  break link\n"
           "Examples:\n"
           "  xln C:\\MSDN\\MSDN120.NT4.COL  C:\\MSDN\\MSDN120.COL\n"
           "  xln -w \"C:\\Program Files\\Far\\Far.exe\" \"C:\\WINDOWS\\Profiles\\All Users\\Start Menu\\Far.lnk\"\n"
           "  xln -n \"C:\\xxx\" \"C:\\WINDOWS\\Profiles\\All Users\\Start Menu\\xxx\"\n"
           "  xln -n -b \"C:\\xxx\"\n"
          );
    exit(0);
} // end print_help()

BOOLEAN
make_shell_link(
    PWCHAR strPathObj,
    PWCHAR _strPathLink,
    PWCHAR SHLinkDescr,
    PWCHAR SHLinkWD,
    PWCHAR SHLinkArg,
    BOOLEAN rel
    )
{

    BOOLEAN bRet = FALSE;
    IShellLinkW* psl;
    IPersistFile* ppf;
    ULONG l;
    PWCHAR strPathLink;

    l = wcslen(_strPathLink);

    if(l>=4 && wcsicmp(_strPathLink+l-4, L".lnk")) {
        l += 4;
        strPathLink = (PWCHAR)GlobalAlloc(GMEM_DISCARDABLE, (l+1)*sizeof(WCHAR));
        if(!strPathLink)
            return FALSE;
        memcpy(strPathLink, _strPathLink, (l-4)*sizeof(WCHAR));
        memcpy(strPathLink+l-4, L".lnk", sizeof(L".lnk"));
    } else {
        strPathLink = _strPathLink;
    }



    if(SUCCEEDED( CoCreateInstance(CLSID_ShellLink,
                            NULL,
                            CLSCTX_INPROC_SERVER,
                            IID_IShellLinkW,
                            (LPVOID*) &psl))
                    )
    {

        if(rel) {
            psl->SetRelativePath(strPathObj, NULL);
        } else {
            psl->SetPath(strPathObj);
        }
        if(SHLinkDescr)
            psl->SetDescription(SHLinkDescr);
        if(SHLinkWD)
            psl->SetWorkingDirectory(SHLinkWD);
        if(SHLinkArg)
            psl->SetArguments(SHLinkArg);

        if (SUCCEEDED(psl->QueryInterface( IID_IPersistFile, (LPVOID *) &ppf)))
        {
            if ( SUCCEEDED ( ppf->Save(strPathLink, TRUE) ) )
                bRet = TRUE;

            ppf->Release();
        }
        psl->Release();
    }

    if(strPathLink != _strPathLink)
        GlobalFree(strPathLink);

    return bRet;
} // end make_shell_link()

BOOLEAN
make_hard_link(
    PWCHAR TargetFN,
    PWCHAR LinkFN
    )
{
    HANDLE          hFile;
    WIN32_STREAM_ID streamId;
    ULONG           bytes_written;
    ULONG           RC;
    LPVOID          lpContext = NULL;
    WCHAR           LinkFullPath[MAX_PATH*2];
    PWCHAR          ptrFN = NULL;
    ULONG           len;

    hFile = CreateFileW(TargetFN,0x100,    // Special privileges
                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                     NULL,
                     OPEN_EXISTING,
                     0,
                     NULL);
    if(hFile == ((HANDLE)-1)) {
        RC = GetLastError();
        printf("Error: cant open target file: %ws (error %#x)\n", TargetFN, RC);
        return FALSE;
    }

    if(!(len = GetFullPathNameW(LinkFN, MAX_PATH*2-1, LinkFullPath, &ptrFN))) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant generate full path for link (error %#x)\n", RC);
        return FALSE;
    }

    LinkFullPath[len] = 0;

    streamId.dwStreamId = BACKUP_LINK;
    streamId.dwStreamAttributes = 0;
    streamId.dwStreamNameSize = 0;
    streamId.Size.HighPart = 0;
    streamId.Size.LowPart = (wcslen(LinkFullPath)+1)*sizeof(WCHAR);

    if (!BackupWrite(hFile, (PUCHAR)&streamId, sizeof(WIN32_STREAM_ID)-4, &bytes_written, FALSE, FALSE, &lpContext)) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant't create link file (error %#x)\n", RC);
        return FALSE;
    }
    if (!BackupWrite(hFile, (PUCHAR)&LinkFullPath[0], streamId.Size.LowPart, &bytes_written, FALSE, FALSE, &lpContext)){
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant't write link info (error %#x)\n", RC);
        exit(1);
    }
    if (!BackupWrite(hFile, NULL, 0, &bytes_written, TRUE, FALSE, &lpContext)){
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant't commit link info (error %#x)\n", RC);
        return FALSE;
    }

    CloseHandle(hFile);
    return TRUE;
} // end make_hard_link()

/*****************************************************
* Thanks to Anton V. Tykhy for NTFS Junction Sources *
*****************************************************/

// At least some SDKs have a buggy value, replacing FILE_ANY_ACCESS with FILE_WRITE_DATA, which
// prevents a recompiled junction from working. Define correct value.
const int _FSCTL_SET_REPARSE_POINT = CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 41, METHOD_BUFFERED, FILE_ANY_ACCESS) ;
const int _FSCTL_DELETE_REPARSE_POINT = CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 43, METHOD_BUFFERED, FILE_ANY_ACCESS) ;

#pragma pack(push, 8)

// Undocumented FSCTL_SET_REPARSE_POINT structure definition
typedef struct {
    DWORD          ReparseTag;
    DWORD          ReparseDataLength;

    WORD           Reserved;
    WORD           ReparseTargetLength;
    WORD           ReparseTargetMaximumLength;
    WORD           Reserved1;
    WCHAR          ReparseTarget[MAX_PATH];
} REPARSE_MOUNTPOINT_DATA_BUFFER, *PREPARSE_MOUNTPOINT_DATA_BUFFER;

#pragma pack(pop)

BOOLEAN
make_ntfs_junction(
    PWCHAR TargetFN,
    PWCHAR LinkFN,
    BOOLEAN SymLink
    )
{
    ULONG           RC;
    REPARSE_MOUNTPOINT_DATA_BUFFER ri;
    DWORD           len;
    HANDLE          hFile;
    WCHAR           TargetFullPath[MAX_PATH*2];
    PWCHAR          ptrFN = NULL;

    CreateDirectoryW(LinkFN, NULL);
    
    hFile = CreateFileW(LinkFN, GENERIC_WRITE, 0, 0, OPEN_EXISTING, 
                            FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS, 0) ;

    if(!hFile || hFile == INVALID_HANDLE_VALUE) {
        RC = GetLastError();
        printf("Error opening reparse point %ws (error %#x)", LinkFN, RC);
        return FALSE;
    }
            
    if(!(len = GetFullPathNameW(TargetFN, MAX_PATH*2-1, TargetFullPath, &ptrFN))) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error: cant generate full path for link (error %#x)\n", RC);
        return FALSE;
    }

    TargetFullPath[len] = 0;

    // Build reparse info
    memset(&ri, 0, sizeof(ri)) ;

    swprintf (ri.ReparseTarget, L"\\??\\%s", TargetFullPath) ;

    ri.ReparseTag                 = SymLink ? 0 : 0xA0000003L; // IO_REPARSE_TAG_MOUNT_POINT winnt ntifs
    ri.ReparseTargetLength        = wcslen(ri.ReparseTarget) * sizeof(WCHAR);
    ri.ReparseTargetMaximumLength = ri.ReparseTargetLength   + sizeof(WCHAR);
    ri.ReparseDataLength          = ri.ReparseTargetLength   + 12 ;

    // Set the link
    if(!DeviceIoControl(hFile, _FSCTL_SET_REPARSE_POINT, &ri, ri.ReparseDataLength + 8, // 2 DWORD header
                                0, 0, &len, 0)) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error setting reparse point to %ws (error %#x)", ptrFN, RC);
        return FALSE;
    }
    printf("%S -> %S\n", LinkFN, ri.ReparseTarget);
    CloseHandle(hFile);
    return TRUE;
} // end make_ntfs_junction()

BOOLEAN
break_ntfs_junction(
    PWCHAR LinkFN,
    BOOLEAN SymLink
    )
{
    ULONG           RC;
    REPARSE_MOUNTPOINT_DATA_BUFFER ri;
    DWORD           len;
    HANDLE          hFile;

    hFile = CreateFileW(LinkFN, GENERIC_WRITE, 0, 0, OPEN_EXISTING, 
                            FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS, 0) ;

    if(!hFile || hFile == INVALID_HANDLE_VALUE) {
        RC = GetLastError();
        printf("Error opening reparse point %ws (error %#x)", LinkFN, RC);
        return FALSE;
    }
            
    // Build reparse info
    memset(&ri, 0, sizeof(ri)) ;

    ri.ReparseTag                 = SymLink ? 0 : 0xA0000003L; // IO_REPARSE_TAG_MOUNT_POINT winnt ntifs
    //ri.ReparseTargetLength        = wcslen(ri.ReparseTarget) * sizeof(WCHAR);
    //ri.ReparseTargetMaximumLength = ri.ReparseTargetLength   + sizeof(WCHAR);
    //ri.ReparseDataLength          = ri.ReparseTargetLength   + 12 ;

    // Set the link
    if(!DeviceIoControl(hFile, _FSCTL_DELETE_REPARSE_POINT, &ri, ri.ReparseDataLength + 8, // 2 DWORD header
                                0, 0, &len, 0)) {
        RC = GetLastError();
        CloseHandle(hFile);
        printf("Error breaking reparse point (error %#x)", RC);
        return FALSE;
    }
    printf("%S -> %S\n", LinkFN, ri.ReparseTarget);
    CloseHandle(hFile);
    return TRUE;
} // end break_ntfs_junction()

int
main (void)
{
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;

    WCHAR* TargetFN = NULL;
    WCHAR* LinkFN = NULL;
    WCHAR* SHLinkDescr = NULL;
    WCHAR* SHLinkWD = NULL;
    WCHAR* SHLinkArg = NULL;

    BOOLEAN rel = FALSE;
    BOOLEAN shell_link = FALSE;
    BOOLEAN ntfs_junc_link = FALSE;
    BOOLEAN ntfs_sym_link = FALSE;
    BOOLEAN ntfs_break_link = FALSE;
    int i;
    int rc;

    printf("NT File System link management utility v3.0 (c) by Alter, 2004-2006\n"
           "Home site http://www.alter.org.ua\n");

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    if(argc < 3) {
        print_help();
    }

    for(i=1; i<argc; i++) {
        if(!argv[i] || argv[i][0] != '-') {
            if(!TargetFN) {
               TargetFN = argv[i];
            } else
            if(!LinkFN) {
               LinkFN = argv[i];
            } else {
                print_help();
                exit(0);
            }
            continue;
        }
        if(!wcscmp(argv[i], L"-h") ||
           !wcscmp(argv[i], L"--help")) {
            print_help();
            exit(0);
        } else
        if(!wcscmp(argv[i], L"-s")/* && !ntfs_junc_link*/) {
            printf("Symbolic links not supported\n");
            exit(0);
        } else
        if(!wcscmp(argv[i], L"-r")) {
            rel = TRUE;
        } else
        if(!wcscmp(argv[i], L"-n")) {
            ntfs_junc_link = TRUE;
        } else
        if(!wcscmp(argv[i], L"-s")) {
            ntfs_sym_link = TRUE;
        } else
        if(!wcscmp(argv[i], L"-b")) {
            ntfs_break_link = TRUE;
        } else
        if(!wcscmp(argv[i], L"-wd")) {
            i++;
            if(i>=argc) {
                print_help();
                exit(0);
            }
            SHLinkWD = argv[i];
        } else
        if(!wcscmp(argv[i], L"-d")) {
            i++;
            if(i>=argc) {
                print_help();
                exit(0);
            }
            SHLinkDescr = argv[i];
        } else
        if(!wcscmp(argv[i], L"-a")) {
            i++;
            if(i>=argc) {
                print_help();
                exit(0);
            }
            SHLinkArg = argv[i];
        } else
        if(!wcscmp(argv[i], L"-w")) {
            shell_link = TRUE;
        } else {
            print_help();
            exit(0);
        }
    }

    if(shell_link) {
        // Create Windows shell link (.lnk)
        CoInitialize(NULL);
        rc = !make_shell_link(TargetFN, LinkFN, SHLinkDescr, SHLinkWD, SHLinkArg, rel);
        CoUninitialize();
    } else {
        // Make NTFS hard or symbolyc link
        if (!Privilege(SE_BACKUP_NAME, TRUE)) {
            printf("Error: insufficient privileges\n");
            return 1;
        }
        if (!Privilege(SE_RESTORE_NAME, TRUE)) {
            printf("Error: insufficient privileges\n");
            return 1;
        }

        if(ntfs_junc_link) {
            if(ntfs_break_link) {
                if(!break_ntfs_junction(TargetFN, ntfs_sym_link)) {
                    return 1;
                }
            } else
            if(!make_ntfs_junction(TargetFN, LinkFN, ntfs_sym_link)) {
                return 1;
            }
        } else {
            if(!make_hard_link(TargetFN, LinkFN)) {
                return 1;
            }
        }
        rc = 0;
    }
    return rc;
} // end main()
